import json
from ...card_utils import (
    CardDecoratorInjector,
    CardRefresher,
    AsyncPeriodicRefresher,
    LineChart,
)

from metaflow.decorators import StepDecorator
from metaflow.cards import Markdown, Table, Artifact
from multiprocessing import Process
from metaflow.metaflow_current import current
from ...utils.disk_usage import usage_collectior
from ...utils.general import unit_convert
import tempfile


class DiskUsageCardRefresher(CardRefresher):

    CARD_ID = "disk_profiler"

    def __init__(self) -> None:
        self._rendered = False
        self._errored = False
        self._line_charts = {}

    def on_error(self, current_card, error_message):
        if isinstance(error_message, FileNotFoundError):
            return
        if isinstance(error_message, json.JSONDecodeError):
            return
        current_card.clear()
        current_card.append(
            Markdown(
                f"## Error: {str(error_message)}",
            )
        )
        current_card.refresh()
        self._errored = True

    def on_startup(self, current_card):
        current_card.append(
            Markdown(
                "# Disk Usage Profiler \n## %s[Attempt:%s]"
                % (current.pathspec, current.retry_count),
            )
        )
        current_card.append(
            Markdown(
                "_waiting for data to appear_",
            )
        )
        current_card.refresh()

    def on_update(self, current_card, data):
        """
        ```
        {
            "/dev/root": [
                {
                    "Filesystem": "/dev/root",
                    "Size": "582G",
                    "Used": "389G",
                    "Avail": "194G",
                    "Use%": "67%",
                    "Mounted": "/",
                    "SizeBytes": 624917741568,
                    "AvailBytes": 20803747840,
                    "Timestamp": "2024-08-13T10:05:20.764734"
                },
                {
                    "Filesystem": "/dev/root",
                    "Size": "582G",
                    "Used": "389G",
                    "Avail": "194G",
                    "Use%": "67%",
                    "Mounted": "/",
                    "SizeBytes": 624917741568,
                    "AvailBytes": 20803747840,
                    "Timestamp": "2024-08-13T10:05:25.766936"
                },
                {
                    "Filesystem": "/dev/root",
                    "Size": "582G",
                    "Used": "389G",
                    "Avail": "194G",
                    "Use%": "67%",
                    "Mounted": "/",
                    "SizeBytes": 624917741568,
                    "AvailBytes": 20803747840,
                    "Timestamp": "2024-08-13T10:05:30.769329"
                },
                {
                    "Filesystem": "/dev/root",
                    "Size": "582G",
                    "Used": "389G",
                    "Avail": "194G",
                    "Use%": "67%",
                    "Mounted": "/",
                    "SizeBytes": 624917741568,
                    "AvailBytes": 20803747840,
                    "Timestamp": "2024-08-13T10:05:35.771963"
                },
            ],
        }
        ```
        """
        if (self._errored and len(data) > 0) or len(self._line_charts) == 0:
            current_card.clear()
            self._errored = False
            self._line_charts = {}

        for fs, usage_list in data.items():
            if fs not in self._line_charts:
                self._line_charts[fs] = LineChart(
                    title=f"Disk Usage for {fs}",
                    xtitle="Timestamp",
                    ytitle="Size",
                    x_name="Timestamp",
                    y_name="Size",
                    x_axis_temporal=True,
                    width=600,
                    height=300,
                )
                current_card.append(self._line_charts[fs])

            self._line_charts[fs].update(
                [
                    {
                        "Timestamp": x["Timestamp"],
                        "Size": unit_convert(x["AvailBytes"], "B", "GB"),
                    }
                    for x in usage_list
                ]
            )
        current_card.refresh()


class DiskUsageProfilerDecorator(StepDecorator, CardDecoratorInjector):

    name = "disk_profiler"

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        self.attach_card_decorator(
            flow, step_name, DiskUsageCardRefresher.CARD_ID, "blank", refresh_interval=2
        )
        self._tempfile = None

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        self._tempfile = tempfile.NamedTemporaryFile(
            prefix="mf_disk_usage", suffix=".json"
        )
        process = Process(target=usage_collectior, args=(self._tempfile.name, 5))
        process.daemon = True
        process.start()

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        def _wrapped_step_func(*args, **kwargs):
            async_disk_usage_refresher = AsyncPeriodicRefresher(
                DiskUsageCardRefresher(),
                updater_interval=5,
                collector_interval=5,
                file_name=self._tempfile.name,
            )

            try:
                async_disk_usage_refresher.start()
                return step_func(*args, **kwargs)
            finally:
                flow._disk_usage_stats = (
                    async_disk_usage_refresher._collector_thread.read()
                )
                async_disk_usage_refresher.stop()

        return _wrapped_step_func



import json
from datetime import datetime


def is_json_serializable(value):
    """Check if value is JSON serializable."""
    try:
        json.dumps(value)
        return True
    except (TypeError, OverflowError):
        return False


def safe_serialize(data):
    def handle_value(val):
        """Converts non-serializable values to serializable format or a placeholder."""
        if isinstance(val, datetime):
            return val.isoformat()
        elif isinstance(val, (dict, list, tuple)):
            return safe_serialize(val)
        elif is_json_serializable(val):
            return val
        else:
            return "<value cannot be JSON serialized>"

    if isinstance(data, dict):
        if len(data) == 0:
            return {}
        return {key: handle_value(val) for key, val in data.items()}
    elif isinstance(data, list):
        return [handle_value(item) for item in data]
    elif isinstance(data, tuple):
        return tuple(handle_value(item) for item in data)
    else:
        return handle_value(data)



# from functools import wraps

import os
from metaflow.metadata.metadata import MetadataProvider
from .core import ModelSerializer, LoadedModels
from ..card_utils import CardDecoratorInjector
from .cards.model_card import ModelListRefresher, ModelsCollector
from ..utils import flowspec_utils

# from .lineage import checkpoint_load_related_metadata, trace_lineage
from ..datastructures import ModelArtifact

# from .cards import CardDecoratorInjector, create_checkpoint_card, null_card
from metaflow.decorators import StepDecorator
from metaflow.metadata import MetaDatum
from typing import List, Dict, Union, Tuple, Optional, Callable, TYPE_CHECKING
from functools import wraps, partial
import tempfile
from metaflow.flowspec import INTERNAL_ARTIFACTS_SET
import json

if TYPE_CHECKING:
    import metaflow


def warning_message(message, logger=None, ts=False, prefix="[@model]"):
    msg = "%s %s" % (prefix, message)
    if logger:
        logger(msg, timestamp=ts, bad=True)


class ModelDecorator(StepDecorator):
    """
    Enables loading / saving of models within a step.


    Parameters
    ----------
    load : Union[List[str],str,List[Tuple[str,Union[str,None]]]], default: None
        Artifact name/s referencing the models/checkpoints to load. Artifact names refer to the names of the instance variables set to `self`.
        These artifact names give to `load` be reference objects or reference `key` string's from objects created by:
        - `current.checkpoint`
        - `current.model`
        - `current.huggingface_hub`

        If a list of tuples is provided, the first element is the artifact name and the second element is the path the artifact needs be unpacked on
        the local filesystem. If the second element is None, the artifact will be unpacked in the current working directory.
        If a string is provided, then the artifact corresponding to that name will be loaded in the current working directory.

    temp_dir_root : str, default: None
        The root directory under which `current.model.loaded` will store loaded models


    MF Add To Current
    -----------------
    model -> metaflow_extensions.obcheckpoint.plugins.machine_learning_utilities.modeling_utils.core.ModelSerializer
        The object used for loading / saving models.
        `current.model` exposes a `save` method to save models and a `load` method to load models.
        `current.model.loaded` exposes the paths to the models loaded via the `load` argument in the @model decorator
        or models loaded via `current.model.load`.

        Usage (Saving a model):
        -------

        ```
        @model
        @step
        def train(self):
            # current.model.save returns a dictionary reference to the model saved
            self.my_model = current.model.save(
                path_to_my_model,
                label="my_model",
                metadata={
                    "epochs": 10,
                    "batch-size": 32,
                    "learning-rate": 0.001,
                }
            )
            self.next(self.test)

        @model(load="my_model")
        @step
        def test(self):
            # `current.model.loaded` returns a dictionary of the loaded models
            # where the key is the name of the artifact and the value is the path to the model
            print(os.listdir(current.model.loaded["my_model"]))
            self.next(self.end)
        ```

        Usage (Loading models):
        -------

        ```
        @step
        def train(self):
            # current.model.load returns the path to the model loaded
            checkpoint_path = current.model.load(
                self.checkpoint_key,
            )
            model_path = current.model.load(
                self.model,
            )
            self.next(self.test)
        ```


        @@ Returns
        -------
        ModelSerializer
            The object used for loading / saving models.
    """

    name = "model"

    defaults = {
        "load": None,  # This can be a list of models to load from artifact references.
        "best_effort_load": False,  # If True, it will ignore missing artifacts and continue.
        "temp_dir_root": None,
    }

    METADATUM_TYPE = "model-registry"

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        self._flow = flow
        self._flow_datastore = flow_datastore
        self._logger = logger
        self._serializer = None
        self.deco_injector = CardDecoratorInjector()
        self.deco_injector.attach_card_decorator(
            flow,
            step_name,
            ModelListRefresher.CARD_ID,
            "blank",
            refresh_interval=2,
        )
        self._collector_thread = None
        # We add to INTERNAL_ARTIFACTS_SET here because the decorator adds internal artifacts to the
        # flow. Adding to INTERNAL_ARTIFACTS_SET avoids having any crashes when `merge_artifacts`
        # is called.
        INTERNAL_ARTIFACTS_SET.add("_task_stored_models")

    def _save_metadata_about_models(self, explicit_model_list: List[Dict]):

        _md_list = []
        self._flow._task_stored_models = explicit_model_list
        # We do `-20` to ensure that we only store the last 20 models.
        for idx, md in enumerate(explicit_model_list[-20:]):
            model_art = ModelArtifact.hydrate(md)
            _md_list.append(
                MetaDatum(
                    field="model-key-%s" % idx,
                    value=model_art.key,
                    type=self.METADATUM_TYPE,
                    tags=[
                        "attempt_id:%s" % str(model_art.attempt),
                    ],
                )
            )
        self._metadata_provider.register_metadata(
            self._runid, self._step_name, self._task_id, _md_list
        )

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata: MetadataProvider,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        self._metadata_provider = metadata
        self._runid = run_id
        self._task_id = task_id
        self._step_name = step_name
        self._setup_current(flow)

    def task_exception(
        self, exception, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        if self._collector_thread is not None:
            self._collector_thread.stop()

    def task_post_step(
        self, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        if self._serializer is None:
            return
        self._save_metadata_about_models(self._serializer._saved_models)
        if self._collector_thread:
            self._collector_thread.stop()

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        from metaflow import current

        self._collector_thread = ModelsCollector(
            ModelListRefresher(current.model.loaded.info),
            interval=1,
        )
        # Wrap the step_func in a function that will write to current.card["checkpoint_info"] the lineage card
        # and then call the step_func.
        def _wrapped_step_func(_collector_thread, *args, **kwargs):
            _collector_thread.start()
            try:
                return step_func(*args, **kwargs)
            finally:
                _collector_thread.stop()

        return partial(_wrapped_step_func, self._collector_thread)

    def _setup_current(self, flow):
        from metaflow import current

        storage_backend = flowspec_utils.resolve_storage_backend(
            run=flow,
        )

        loaded_models = LoadedModels(
            storage_backend=storage_backend,
            flow=flow,
            logger=self._logger,
            artifact_references=[]
            if self.attributes["load"] is None
            else self.attributes["load"],
            best_effort=self.attributes["best_effort_load"],
            temp_dir_root=self.attributes["temp_dir_root"],
        )

        self.loaded_models_data = loaded_models.info

        serializer = ModelSerializer(
            pathspec=current.pathspec,
            attempt=current.retry_count,
            storage_backend=storage_backend,
        )
        serializer._set_loaded_models(loaded_models)
        current._update_env(
            {
                "model": serializer,
            }
        )
        self._serializer = serializer



from typing import Union, Optional, TYPE_CHECKING
from metaflow.metadata import MetaDatum
from ..datastructures import CheckpointArtifact
from .core import CheckpointReferenceResolver

if TYPE_CHECKING:
    import metaflow


def checkpoint_load_related_metadata(checkpoint: CheckpointArtifact, current_attempt):
    data = {
        "checkpoint-load-key": checkpoint.key,
        "checkpoint-load-pathspec": checkpoint.pathspec,
        "checkpoint-load-attempt": checkpoint.attempt,
        "checkpoint-load-name": checkpoint.name,
    }
    md = []
    for k, v in data.items():
        md.append(
            MetaDatum(
                field=k,
                value=v,
                type="checkpoint",
                tags=[
                    "attempt_id:%s" % str(current_attempt),
                ],
            )
        )
    return md


def trace_lineage(flow, checkpoint: CheckpointArtifact):
    """
    Trace the lineage of the checkpoint by tracing the previous paths.
    """
    from metaflow import Task

    def _loaded_parent_key(
        _checkpoint: CheckpointArtifact, _flow, parents=[], max_depth=20
    ):
        try:
            metadata = Task(
                _checkpoint.pathspec, attempt=_checkpoint.attempt
            ).metadata_dict
        except AttributeError:
            return parents  # No parents were discovered.
        if "checkpoint-load-key" not in metadata:
            return parents
        data = CheckpointReferenceResolver.from_key(
            _flow, metadata.get("checkpoint-load-key")
        )
        if data is None:
            return parents
        if data.key == _checkpoint.key:
            # This means that we are iterating over the same checkpoint again and again
            # because the checkpoint-load-key of the task is bringing it back to the same task.
            # Ideally this shouldn't happen unless there is bug in the `checkpoint-load-key` metadata
            # setter.
            return parents

        parents.append(data)
        if max_depth == len(parents):
            return parents

        return _loaded_parent_key(data, _flow, parents=parents, max_depth=max_depth)

    return _loaded_parent_key(checkpoint, flow, parents=[])



from collections import defaultdict
import subprocess
import json
import time
from metaflow._vendor import click
import signal
import sys
from datetime import datetime

FILE_PATH = __file__


def signal_handler(sig, frame):
    sys.exit(0)


signal.signal(signal.SIGINT, signal_handler)


def parse_size(size_str):
    """
    Convert human-readable size (e.g., '1K', '1M', '1G', '1T') to bytes.
    """
    size_str = size_str.strip()
    if size_str.endswith("K"):
        return int(float(size_str[:-1]) * 1024)
    elif size_str.endswith("M"):
        return int(float(size_str[:-1]) * 1024**2)
    elif size_str.endswith("G"):
        return int(float(size_str[:-1]) * 1024**3)
    elif size_str.endswith("T"):
        return int(float(size_str[:-1]) * 1024**4)
    elif size_str.endswith("P"):
        return int(float(size_str[:-1]) * 1024**5)
    else:
        return int(size_str)  # Assume it's already in bytes if no suffix


def disk_usage_to_json():
    try:
        # Run the df command
        df_output = subprocess.check_output(["df", "-h"], text=True)

        # Split output into lines
        lines = df_output.strip().split("\n")

        # Extract headers
        headers = lines[0].split()

        # Initialize list to hold disk usage information
        disk_usage_info = []

        # Parse each line of df output after the header
        for line in lines[1:]:
            columns = line.split()
            # Map each column to its header
            usage_dict = dict(zip(headers, columns))

            # Convert the 'Size' column to bytes and add it as a new key
            if "Size" in usage_dict:
                usage_dict["SizeBytes"] = parse_size(usage_dict["Size"])

            if "Avail" in usage_dict:
                usage_dict["AvailBytes"] = parse_size(usage_dict["Avail"])

            disk_usage_info.append(usage_dict)

        # Convert the list to JSON format
        return disk_usage_info

    except subprocess.CalledProcessError as e:
        return {"error": str(e)}


class UsageCounter:
    def __init__(self):
        self._fs_usage = defaultdict(list)

    def update(self, usage_dict):
        """
        `usage_dict` looks like. This thing will update the usage counter for each filesystem.
            {
                "Filesystem": "/dev/root",
                "Size": "582G",
                "Used": "389G",
                "Avail": "194G",
                "Use%": "67%",
                "Mounted": "/",
                "SizeBytes": 624917741568,
                "AvailBytes": 20803747840,
            },

        """
        self._fs_usage[
            usage_dict["Filesystem"] + "(Mounted: %s)" % usage_dict["Mounted"]
        ].append({**usage_dict, "Timestamp": datetime.now().isoformat()})

    def dump(self):
        return json.dumps(self._fs_usage, indent=4)


def usage_collectior(file, polling_interval):
    usage_counter = UsageCounter()
    try:
        while True:
            disk_usage_dict = disk_usage_to_json()
            if "error" in disk_usage_dict:
                print("Error getting disk usage information", file=sys.stderr)
                print(disk_usage_dict["error"], file=sys.stderr)
            else:
                for usage_dict in disk_usage_dict:
                    usage_counter.update(usage_dict)
                if file:
                    with open(file, "w") as f:
                        f.write(usage_counter.dump())
                else:
                    print(usage_counter.dump())
                time.sleep(polling_interval)
    except KeyboardInterrupt:
        pass


@click.command()
@click.option("--file", "-f", type=click.Path(), help="Output file to write JSON to")
@click.option(
    "--polling_interval",
    "--interval",
    "-i",
    type=int,
    default=5,
    help="Polling interval in seconds",
)
def run(file, polling_interval):
    usage_collectior(file, polling_interval)


# Example usage
if __name__ == "__main__":
    run()



import hashlib
import shutil
from ..checkpoints.decorator import (
    CheckpointDecorator,
    CurrentCheckpointer,
    warning_message,
)

HUGGINGFACE_HUB_ROOT_PREFIX = "mf.huggingface_hub"


def download_model_from_huggingface(**kwargs):
    import huggingface_hub
    from glob import glob
    import os

    try:
        huggingface_hub.snapshot_download(**kwargs)
        pass
    except Exception as e:
        raise e


class HuggingfaceRegistry:
    _checkpointer: CurrentCheckpointer = None

    def __init__(self, logger) -> None:
        self._logger = logger

    def _set_checkpointer(self, checkpointer: CurrentCheckpointer):
        self._checkpointer = checkpointer
        self._checkpointer._default_checkpointer._checkpointer.set_root_prefix(
            HUGGINGFACE_HUB_ROOT_PREFIX
        )

    def _cache_name(self, name):
        return hashlib.md5(name.encode()).hexdigest()[:10]

    def _warn(self, message):
        warning_message(
            message,
            self._logger,
            prefix="[@huggingface_hub]",
        )

    def _load_or_cache_model(self, **kwargs):
        from metaflow import current

        repo_name = kwargs["repo_id"]
        repo_type = kwargs.get("repo_type", "model")
        force_download = kwargs.get("force_download", False)
        chckpt_name = self._cache_name("{}/{}".format(repo_type, repo_name))
        chckpts = list(self._checkpointer.list(name=chckpt_name, within_task=False))
        if len(chckpts) > 0 and not force_download:
            return chckpts[0]

        _kwargs = kwargs.copy()
        _kwargs["local_dir"] = self._checkpointer.directory
        _kwargs["local_dir_use_symlinks"] = False
        self._warn(
            "Downloading %s from huggingface" % repo_name,
        )
        download_model_from_huggingface(**_kwargs)
        self._warn(
            "Downloaded %s from huggingface. Saving checkpoint" % repo_name,
        )
        chckpt_ref = self._checkpointer.save(
            name=chckpt_name,
            metadata={
                "repo_id": repo_name,
                "registry": "huggingface",
                "repo_type": repo_type,
            },
            # We set this statically to files here because
            # it will be a lot more performant than tar mode.
            storage_format="files",
        )
        self._warn(
            "huggingface checkpoint for %s saved" % repo_name,
        )
        # wipe the directory so that it's unpolluted for another function call.
        shutil.rmtree(self._checkpointer.directory)
        return chckpt_ref

    def snapshot_download(self, **kwargs):
        if "repo_id" not in kwargs:
            raise ValueError("repo_id is required for snapshot_download")
        return self._load_or_cache_model(**kwargs)


class HuggingfaceHubDecorator(CheckpointDecorator):

    defaults = {}

    name = "huggingface_hub"

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        self._flow_datastore = flow_datastore
        self._logger = logger
        self._chkptr = None
        self._collector_thread = None
        self._registry = HuggingfaceRegistry(logger)

    def _resolve_settings(self):
        return {"scope": "namespace", "load_policy": "none"}

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        super().task_pre_step(
            step_name,
            task_datastore,
            metadata,
            run_id,
            task_id,
            flow,
            graph,
            retry_count,
            max_user_code_retries,
            ubf_context,
            inputs,
        )

    def _setup_current(self):
        from metaflow import current

        self._registry._set_checkpointer(self._chkptr)
        current._update_env({"huggingface_hub": self._registry})

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        return step_func

    def task_post_step(
        self, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        self._chkptr.cleanup()

    def task_exception(
        self, exception, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        self._chkptr.cleanup()



from re import M
from typing import Union

# from tarfile import SUPPORTED_TYPES
from typing import Optional, List
from datetime import datetime
from .exceptions import (
    KeyNotCompatibleWithObjectException,
    KeyNotCompatibleException,
    IncompatibleObjectTypeException,
)
from .datastore.task_utils import init_datastorage_object


class MetaflowDataArtifactReference:

    _MODULE_IMPORTS = {}

    TYPE = None

    REQUIRED_FIELDS = {
        "url": str,
        "key": str,
        "pathspec": str,
        "attempt": int,
        "created_on": str,
        "type": str,
    }

    OPTIONAL_FIELDS = {
        "metadata": dict,
        "name": str,
        "size": int,
    }

    @property
    def size(self):
        return self._values.get("size", None)

    # PROPERTIES :
    @property
    def url(self):
        return self._values["url"]

    @property
    def key(self):
        return self._values["key"]

    @property
    def pathspec(self):
        return self._values.get("pathspec")

    @property
    def attempt(self):
        return self._values.get("attempt")

    @property
    def created_on(self):
        return self._values["created_on"]

    @property
    def metadata(self):
        return self._values.get("metadata")

    def __init__(self, **kwargs):
        self.validate(kwargs)
        self._values = kwargs

    def validate(self, data):
        for key, value in data.items():
            if key in self.REQUIRED_FIELDS and not isinstance(
                value, self.REQUIRED_FIELDS[key]
            ):
                raise ValueError(
                    f"Value {value} for key {key} not of type {self.REQUIRED_FIELDS[key]}"
                )

        misisng_keys = set(self.REQUIRED_FIELDS.keys()) - set(
            data.keys()
        )  # keys in schema but not in data
        if misisng_keys:
            # Missing keys for required fields is not fine
            raise ValueError(f"Missing keys {misisng_keys} for required fields")

        return True

    @classmethod
    def from_dict(cls, data) -> Union["ModelArtifact", "CheckpointArtifact"]:
        return Factory.from_dict(data)

    @classmethod
    def hydrate(
        cls,
        data: Union["ModelArtifact", "CheckpointArtifact", dict],
    ):
        if isinstance(data, dict):
            return cls.from_dict(data)
        return data

    def to_dict(self):
        values = self._values.copy()
        values["type"] = self.TYPE
        return values

    # ! INHERIT AND OVERRIDE (Private from user)
    def _load(self, storage_backend, local_path):
        raise NotImplementedError

    # ! INHERIT AND OVERRIDE (Private from user)
    @classmethod
    def _valid_reference_key(cls, reference_key):
        raise NotImplementedError

    @classmethod
    def _load_from_key(cls, key, local_path, storage_backend):
        raise NotImplementedError

    @classmethod
    def _load_metadata_from_key(cls, key, storage_backend):
        raise NotImplementedError


class ModelArtifact(MetaflowDataArtifactReference):

    OPTIONAL_FIELDS = {
        "metadata": dict,
        "serializer": Optional[str],
        "source": Optional[
            str
        ],  # Did it come from metaflow Task from something liked HF-HUB, etc.
        "lineage": Optional[List],
        "url": str,
        "key": str,
        "storage_format": str,
        "blob": bytes,  # bytes blob
        "label": str,
        "size": int,  # In bytes
    }

    REQUIRED_FIELDS = {
        "pathspec": str,
        "attempt": int,
        "created_on": str,
        "type": str,
        "model_uuid": str,
    }

    TYPE = "model"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def blob(self):
        return self._values.get("blob", None)

    @property
    def uuid(self):
        return self._values.get("model_uuid", None)

    @property
    def serializer(self):
        return self._values.get("serializer", None)

    @property
    def source(self):
        return self._values.get("source", None)

    @property
    def storage_format(self):
        return self._values.get("storage_format", None)

    @classmethod
    def create(
        cls,
        pathspec=None,
        attempt=None,
        key=None,
        url=None,
        model_uuid=None,
        metadata=None,
        storage_format=None,
        source=None,
        serializer=None,
        label=None,
    ):
        return cls(
            pathspec=pathspec,
            attempt=attempt,
            key=key,
            created_on=datetime.now().isoformat(),
            type=cls.TYPE,
            url=url,
            model_uuid=model_uuid,
            metadata=metadata,
            source=source,
            storage_format=storage_format,
            serializer=serializer,
            label=label,
        )

    def _set_blob(self, blob: bytes):
        self._values["blob"] = blob

    def _load(self, storage_backend, local_path):
        from .modeling_utils.core import _load_model

        _load_model(storage_backend, self.key, local_path)

    @classmethod
    def _valid_reference_key(cls, reference_key):
        from .modeling_utils.model_storage import ModelDatastore

        try:
            ModelDatastore.decompose_key(reference_key)
            return True
        except KeyNotCompatibleWithObjectException:
            return False

    @classmethod
    def _load_from_key(cls, key, local_path, storage_backend):
        from .modeling_utils.core import _load_model

        _load_model(storage_backend, model_key=key, path=local_path)

    @classmethod
    def _load_metadata_from_key(cls, key, storage_backend) -> "ModelArtifact":
        from .modeling_utils.core import _load_model_metadata

        return cls(
            **_load_model_metadata(
                storage_backend,
                model_key=key,
            )
        )


class CheckpointArtifact(MetaflowDataArtifactReference):

    TYPE = "checkpoint"

    REQUIRED_FIELDS = {
        "url": str,
        "key": str,
        "created_on": str,
        "type": str,
        # This means weather the checkpoint was created within Metaflow context or outside Metaflow context.
        # `creation_context` can be either `task` or `non-task`
    }

    OPTIONAL_FIELDS = {
        "creation_context": str,
        "pathspec": str,
        "attempt": int,
        "metadata": dict,
        "name": str,
        "version_id": str,
        "storage_format": str,
    }

    @property
    def storage_format(self):
        return self._values.get("storage_format", None)

    @property
    def version_id(self):
        return self._values.get("version_id", None)

    @property
    def name(self):
        return self._values.get("name")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _load(self, storage_backend, local_path):
        # TODO [POST RELEASE]: Check if we can move this somewhere else.
        # TODO : The connective tissue with checkpointer is not designed optimally.
        # Need to fix the design of checkpointer and constructors to figure
        # the best means of using the abstractions.
        from .checkpoints.core import Checkpointer

        _checkpointer = Checkpointer._from_checkpoint_and_storage_backend(
            self, storage_backend
        )
        _checkpointer._load_checkpoint(
            local_path=local_path,
            version_id=self.version_id,
            name=self.name,
            storage_format=self.storage_format,
        )

    @classmethod
    def _valid_reference_key(cls, reference_key):
        from .checkpoints.checkpoint_storage import CheckpointDatastore

        try:
            CheckpointDatastore.decompose_key(reference_key)
            return True
        except KeyNotCompatibleWithObjectException:
            return False

    @classmethod
    def _load_from_key(cls, key, local_path, storage_backend):
        from .checkpoints.constructors import load_checkpoint

        load_checkpoint(key, local_path)

    @classmethod
    def _load_metadata_from_key(cls, key, storage_backend):
        from .checkpoints.core import Checkpointer
        from .checkpoints.checkpoint_storage import CheckpointDatastore

        key_decomp = CheckpointDatastore.decompose_key(key)
        _checkpointer = Checkpointer._from_key(key)
        return cls(
            **_checkpointer.load_metadata(
                version_id=key_decomp.version_id,
                name=key_decomp.name,
            )
        )


class Factory:

    SUPPORTED_TYPES = [
        ModelArtifact.TYPE,
        CheckpointArtifact.TYPE,
    ]

    @classmethod
    def hydrate(cls, data):
        if isinstance(data, dict):
            return cls.from_dict(data)
        elif type(data) in [
            ModelArtifact,
            CheckpointArtifact,
        ]:
            return data
        else:
            raise ValueError("Data is not a dict or an instance of Artifact")

    @classmethod
    def from_dict(cls, data):
        if "type" not in data or data["type"] not in cls.SUPPORTED_TYPES:
            raise IncompatibleObjectTypeException(
                "Object type %s not in supported types: %s"
                % (data.get("type") or str(data), cls.SUPPORTED_TYPES)
            )
        if data["type"] == ModelArtifact.TYPE:
            return ModelArtifact(**data)
        if data["type"] == CheckpointArtifact.TYPE:
            return CheckpointArtifact(**data)

    @classmethod
    def load(cls, data, local_path, storage_backend):
        # `data` here is the dictionary/python object
        art = cls.hydrate(data)
        art._load(storage_backend, local_path)

    @classmethod
    def object_type_from_key(cls, reference_key):
        for obj_type in [ModelArtifact, CheckpointArtifact]:
            if obj_type._valid_reference_key(reference_key):
                return obj_type
        return None

    @classmethod
    def load_from_key(cls, key_object, local_path, storage_backend):
        obj_type = cls.object_type_from_key(key_object)
        if obj_type is None:
            raise KeyNotCompatibleException(
                key_object,
                supported_types=", ".join(
                    [ModelArtifact.TYPE, CheckpointArtifact.TYPE]
                ),
            )
        obj_type._load_from_key(key_object, local_path, storage_backend)

    @classmethod
    def load_metadata_from_key(
        cls, key_object, storage_backend
    ) -> Union[CheckpointArtifact, ModelArtifact]:
        obj_type = cls.object_type_from_key(key_object)
        if obj_type is None:
            raise KeyNotCompatibleException(
                key_object,
                supported_types=", ".join(
                    [ModelArtifact.TYPE, CheckpointArtifact.TYPE]
                ),
            )
        return obj_type._load_metadata_from_key(key_object, storage_backend)


# This will be the core user facing function that will help load stuff
def load_model(
    reference: Union[str, MetaflowDataArtifactReference, dict],
    path: str,
):
    if reference is None:
        raise ValueError("`load_model` requires a reference")
    if path is None:
        raise ValueError("`load_model` requires a path to load the model")

    storage_backend = init_datastorage_object()
    if type(reference) == dict or isinstance(reference, MetaflowDataArtifactReference):
        Factory.load(
            Factory.hydrate(reference),
            path,
            storage_backend,
        )
    elif type(reference) == str:
        Factory.load_from_key(reference, path, storage_backend)



import json
import time
from typing import Dict, List, Union
from ...card_utils import (
    CardRefresher,
    UpadateableTable,
)
from ...datastructures import CheckpointArtifact, ModelArtifact
from ...utils.general import unit_convert
from metaflow.cards import Markdown, Table, Artifact, VegaChart
from datetime import datetime, timedelta
from threading import Thread, Event


import json
from datetime import datetime


# Helper function to format datetime
def format_datetime(iso_str):
    # Parse the ISO 8601 string to a datetime object
    dt = datetime.fromisoformat(iso_str)
    # Format the datetime object to a more readable string
    return dt.strftime("%B %d, %Y, %H:%M:%S")


def human_readable_date(date):
    return date.strftime("%Y-%m-%d %H:%M:%S")


def determine_nice_value(time_range_seconds):
    """
    Function to determine the 'nice' value based on the time range in seconds.
    """
    if time_range_seconds <= 60 * 30:  # less than 30 mins
        return "second", "%H:%M:%S"
    elif time_range_seconds <= 3600:  # less than an hour
        return "minute", "%H:%M"
    elif time_range_seconds <= 86400:  # less than a day
        return "hour", "%H:%M"
    elif time_range_seconds <= 604800:  # less than a week
        return "day", "%Y-%m-%d"
    elif time_range_seconds <= 2592000:  # less than a month
        return "week", "%Y-%m-%d"
    elif time_range_seconds <= 31536000:  # less than a year
        return "month", "%Y-%m"
    else:
        return "year", "%Y"


def _get_table_row(object):
    if object["type"] == ModelArtifact.TYPE:
        return [
            Markdown(object["model_uuid"]),
            Artifact(object["type"]),
            Markdown(format_datetime(object["created_on"])),
            Markdown(_derive_appropriate_size(object["size"])),
            Artifact(object["metadata"]),
            Artifact(object["key"]),
        ]
    elif object["type"] == CheckpointArtifact.TYPE:
        return [
            Markdown(
                "%s.%s.%s"
                % (
                    object["attempt"],
                    object["name"],
                    object["version_id"],
                )
            ),
            Artifact(object["type"]),
            Markdown(format_datetime(object["created_on"])),
            Markdown(_derive_appropriate_size(object["size"])),
            Artifact(object["metadata"]),
            Artifact(object["key"]),
        ]
    return []


def _make_loaded_models_table(loaded_models: Dict[str, Dict]):
    data = []
    headers = [
        "Artifact Name",
        "ID",
        "Artifact Type",
        "Created On",
        "Size",
        "Metadata",
        "Key",
    ]

    for key, object in loaded_models.items():
        x = _get_table_row(object)
        x.insert(0, Markdown(key))
        data.append(x)

    return UpadateableTable(
        headers=headers,
        data=data,
    )


class ModelListRefresher(CardRefresher):

    CARD_ID = "model_card"

    TABLE_HEADERS = [
        "Name",
        "Created On",
        "Size",
        "Metadata",
        # "Key",
    ]

    def __init__(
        self,
        loaded_models: Dict[str, Dict],
    ) -> None:
        self._rendered = False
        self._errored = False
        self._table = None
        self._saved_models = {}
        self._loaded_models = loaded_models

        from metaflow import current

        self.current = current

    def on_error(self, current_card, error_message):
        if isinstance(error_message, FileNotFoundError):
            return
        if isinstance(error_message, json.JSONDecodeError):
            return
        current_card.clear()
        current_card.append(
            Markdown(
                f"## Error: {str(error_message)}",
            )
        )
        current_card.refresh()
        self._errored = True
        self._rendered = False

    def _header_components(self):
        x = [
            Markdown(
                "# Models \n **Task %s [Attempt:%s]**"
                % (self.current.pathspec, self.current.retry_count),
            ),
            Markdown("## Loaded Models"),
        ]
        if len(self._loaded_models) > 0:
            x.append(_make_loaded_models_table(self._loaded_models))
        else:
            x.append(Markdown("_No models loaded_"))
        return x

    def _footer_components(self):
        return []

    def on_startup(self, current_card):
        current_card.extend(self._header_components())
        current_card.extend(self._footer_components())
        current_card.refresh()

    def _make_table_objects(self, data_object):
        keys_going_in_table = []
        for model in data_object:
            _model = ModelArtifact.from_dict(model)
            if _model.key in self._saved_models:
                continue
            self._saved_models[_model.key] = [
                Markdown(str(_model.uuid)),
                Markdown(format_datetime(str(_model.created_on))),
                Markdown(
                    _derive_appropriate_size(_model.size),
                ),
                Artifact(_model.metadata),
                # Markdown(_model.key),
            ]
            keys_going_in_table.append(_model.key)
        return keys_going_in_table

    def first_time_render(self, current_card, saved_models, force_refresh=False):
        current_card.clear()
        current_card.extend(self._header_components())
        keys_going_in_table = self._make_table_objects(saved_models)
        current_card.append(Markdown("## New Models Created"))
        if len(keys_going_in_table) == 0:
            current_card.append(Markdown("_No new models created_"))
            current_card.refresh()
            return

        self._table = UpadateableTable(
            data=[self._saved_models[key] for key in keys_going_in_table],
            headers=self.TABLE_HEADERS,
        )
        current_card.append(self._table)

        current_card.extend(self._footer_components())
        current_card.refresh(force=force_refresh)
        self._rendered = True

    def data_update(self, current_card, saved_models):
        keys_going_in_table = self._make_table_objects(saved_models)
        if len(keys_going_in_table) == 0:
            return
        for key in keys_going_in_table:
            self._table.update(self._saved_models[key])

        current_card.refresh()

    def on_update(self, current_card, saved_models):
        if not self._rendered:
            self.first_time_render(current_card, saved_models, force_refresh=False)
        else:
            self.data_update(current_card, saved_models)

    def on_final(self, current_card, data_object):
        self._saved_models = {}
        self._rendered = False
        self.first_time_render(current_card, data_object, force_refresh=True)


def _derive_appropriate_size(size):
    if unit_convert(size, "B", "MB") < 1:
        return "%s KB" % str(unit_convert(size, "B", "KB"))
    elif unit_convert(size, "B", "GB") < 1:
        return "%s MB" % str(unit_convert(size, "B", "MB"))
    else:
        return "%s GB" % str(unit_convert(size, "B", "GB"))


class ModelsCollector(Thread):
    def __init__(self, refresher: CardRefresher, interval=1):
        super().__init__()
        from metaflow import current

        self.current = current
        self._interval = interval
        self._exit_event = Event()
        self._refresher = refresher

    def collect(self):
        return self.current.model._saved_models

    def final_update(self):
        current_card = self.current.card[self._refresher.CARD_ID]
        data = self.collect()
        if len(data) == 0:
            return
        self._refresher.on_final(current_card, data)

    def run_update(self):
        current_card = self.current.card[self._refresher.CARD_ID]
        data = self.collect()
        if len(data) == 0:
            return
        self._refresher.on_update(current_card, data)

    def run(self):
        if self._refresher.CARD_ID is None:
            raise ValueError("CARD_ID must be defined")
        current_card = self.current.card[self._refresher.CARD_ID]
        self._refresher.on_startup(current_card)
        while self._exit_event.is_set() is False:
            self.run_update()
            time.sleep(self._interval)

    def stop(self):
        if not self._exit_event.is_set():
            self._exit_event.set()
            # We expose a `final_update` so that the card can be
            # called with a `force` update so that the new card
            # is rendered when the thread is stopped.
            self.final_update()
            self.join()



from metaflow.exception import MetaflowException


class TODOException(MetaflowException):
    headline = "TODOException"

    def __init__(self, message):
        super(TODOException, self).__init__(message)


class KeyNotFoundError(MetaflowException):
    def __init__(self, url):
        msg = "Key not found: %s" % url
        self._url = url
        super().__init__(msg)


class KeyNotCompatibleException(MetaflowException):
    headline = "Object key incompatible with any supported objects"

    def __init__(self, key, supported_types):
        msg = "Key %s is not compatible with any of the supported types: %s" % (
            key,
            supported_types,
        )
        self._key = key
        super().__init__(msg)


class KeyNotCompatibleWithObjectException(MetaflowException):
    def __init__(self, key, store, message=None):
        msg = "Key %s is not compatible with object of type `%s`" % (
            key,
            store,
        )
        if message:
            msg += "\n" + message
        self._key = key
        super().__init__(msg)


class IncompatibleObjectTypeException(MetaflowException):
    headline = "Object Type Incompatible"

    def __init__(self, message):
        super().__init__(message)



"""
This file contains methods injected into the `DataStoreStorage` class
"""
from metaflow.plugins.datastores.s3_storage import S3, ARTIFACT_LOCALROOT
from metaflow.plugins.datastores.gs_storage import (
    handle_executor_exceptions,
    as_completed,
)
from metaflow.plugins.datatools.s3 import S3PutObject
from metaflow.plugins.datastores.local_storage import LocalStorage
import shutil
import os


def save_files_s3(self, key_path_tuples, overwrite=False):
    with S3(
        s3root=self.datastore_root,
        tmproot=ARTIFACT_LOCALROOT,
        external_client=self.s3_client,
    ) as s3:
        s3.put_files(
            [
                S3PutObject(
                    key=key,
                    path=path,
                )
                for key, path in key_path_tuples
            ],
            overwrite=overwrite,
        )


def save_file_s3(self, key, path, overwrite=False):
    with S3(
        s3root=self.datastore_root,
        tmproot=ARTIFACT_LOCALROOT,
        external_client=self.s3_client,
    ) as s3:
        s3.put_files(
            [
                S3PutObject(
                    key=key,
                    path=path,
                )
            ],
            overwrite=overwrite,
        )


@handle_executor_exceptions
def save_files_gcp_or_azure(self, key_path_tuples, overwrite=False):
    futures = []
    for key, path in key_path_tuples:
        futures.append(
            self._executor.submit(
                self.root_client.save_bytes_single,
                (key, path, {"save_file_single": True}),
                overwrite=overwrite,
            )
        )
    for future in as_completed(futures):
        future.result()


@handle_executor_exceptions
def save_file_gcp_or_azure(self, key, path, overwrite=False):
    futures = []
    futures.append(
        self._executor.submit(
            self.root_client.save_bytes_single,
            (key, path, {"save_file_single": True}),
            overwrite=overwrite,
        )
    )
    for future in as_completed(futures):
        future.result()


def save_file_local(self, key, path, overwrite=False):

    full_path = self.full_uri(key)
    if not overwrite and os.path.exists(full_path):
        return
    LocalStorage._makedirs(os.path.dirname(full_path))
    shutil.copy(path, full_path)


def save_files_local(self, key_path_tuples, overwrite=False):
    for key, path in key_path_tuples:
        save_file_local(self, key, path, overwrite=overwrite)


STORAGE_INJECTIONS_SINGLE_FILE_SAVE = {
    "s3": save_file_s3,
    "gs": save_file_gcp_or_azure,
    "azure": save_file_gcp_or_azure,
    "local": save_file_local,
}

STORAGE_INJECTIONS_MULTIPLE_FILE_SAVE = {
    "s3": save_files_s3,
    "gs": save_files_gcp_or_azure,
    "azure": save_files_gcp_or_azure,
    "local": save_files_local,
}



# THIS FILE IS ALSO USED (Figure where is the best place to put it.)
from typing import Union, Optional, TYPE_CHECKING

from metaflow.exception import MetaflowException
from metaflow.datastore import FlowDataStore
from .core import resolve_root


if TYPE_CHECKING:
    import metaflow


class UnresolvableDatastoreException(MetaflowException):
    pass


def _get_flow_datastore(task):
    flow_name = task.pathspec.split("/")[0]
    # Resolve datastore type
    ds_type = None
    # We need to set the correct datastore root here so that
    # we can ensure that the card client picks up the correct path to the cards

    meta_dict = task.metadata_dict
    ds_type = meta_dict.get("ds-type", None)

    if ds_type is None:
        raise UnresolvableDatastoreException(task)

    ds_root = meta_dict.get("ds-root", None)

    if ds_root is None:
        ds_root = resolve_root(ds_type)

    # Delay load to prevent circular dep
    from metaflow.plugins import DATASTORES

    storage_impl = [d for d in DATASTORES if d.TYPE == ds_type][0]
    return FlowDataStore(
        flow_name=flow_name,
        environment=None,
        storage_impl=storage_impl,
        # ! ds root cannot be none otherwise `list_content`
        # ! method fails in the datastore abstraction.
        ds_root=ds_root,
    )


def init_datastorage_object():
    from metaflow.plugins import DATASTORES
    from metaflow.metaflow_config import DEFAULT_DATASTORE

    storage_impl = [d for d in DATASTORES if d.TYPE == DEFAULT_DATASTORE][0]
    return storage_impl(storage_impl.get_datastore_root_from_config(print))


def resolve_storage_backend(pathspec: Union[str, "metaflow.Task"] = None):
    from metaflow.client.core import Task

    if isinstance(pathspec, Task):
        return _get_flow_datastore(pathspec)._storage_impl
    elif isinstance(pathspec, str):
        if len(pathspec.split("/")) != 4:
            raise ValueError("Pathspec is not of the correct format.")
        return _get_flow_datastore(Task(pathspec))._storage_impl
    else:
        raise ValueError(
            "Pathspec is of invalid type. It should be either a string or a Task object but got %s"
            % type(pathspec)
        )



import os
import tarfile
from typing import Union, List
import sys
from pathlib import Path
from metaflow.exception import MetaflowException
from io import BytesIO


class TarballCreationError(MetaflowException):
    headline = "Error creating tarball"

    def __init__(self, message, error):
        super().__init__(message)
        self.error = error


def _get_size_of_path(path):
    root_directory = Path(path)
    return sum(f.stat().st_size for f in root_directory.glob("**/*") if f.is_file())


def warning_message(
    message, logger=None, ts=False, prefix="[@checkpoint][tarball-creator]"
):
    msg = "%s %s" % (prefix, message)
    if logger:
        logger(msg, timestamp=ts, bad=True)
    else:
        print(msg, file=sys.stderr)


def os_error_safe_file_adding(tar, file_path, arcname):
    try:
        tar.add(file_path, arcname=arcname)
    except OSError as e:
        warning_message(
            f"Error adding {file_path} to tarball. Error: {e}",
            prefix="[@checkpoint][tarball-creator]",
        )


def _prepare_arc_names_and_sources(
    source_paths: Union[str, List[str]],
):
    _paths = source_paths
    if isinstance(source_paths, str):
        _paths = [source_paths]
    # When we save tarballs we try to preserve the directory
    # structure based on when the `checkpoint` method is called.
    # The _resolve_arc_name method will ensure that directory
    # structure is fully preserved when only a single directory
    # is passed to the `create_tarball` method.
    _arcnames = _resolve_arc_name(_paths)
    return _paths, _arcnames


def _save_path_and_arcs_to_tar(tar, _paths, _arcnames, strict=False):
    for sp, arcn in zip(_paths, _arcnames):
        # when we create these tar balls. There is a non-zero possibility that
        # the file might get changed/removed while adding it to the tar-archive.
        # This is because many of these checkpointing scenarios take place in a
        # multiprocessing environment. So having a safety check and also propagating
        # that information in the checkpoint data is a good idea.
        # Ideally it would be a great idea if the `save` function for a checkpointer
        # can take an option that says that TAR ball creation is best-effort OR
        # Fully safe.
        if not strict:
            os_error_safe_file_adding(tar, sp, arcn)
        else:
            try:
                tar.add(sp, arcname=arcn)
            except OSError as e:
                raise TarballCreationError(
                    f"Error adding {sp} to tarball. Resulted in an OSError: {e}", e
                )


def create_tarball_in_memory(
    source_paths: Union[str, List[str]], compression_method="gz", strict=False
):
    _file = BytesIO()
    _paths, _arcnames = _prepare_arc_names_and_sources(source_paths)
    write_string = "w"
    if compression_method is not None:
        write_string = "w:%s" % compression_method
    with tarfile.open(fileobj=_file, mode=write_string) as tar:
        _save_path_and_arcs_to_tar(tar, _paths, _arcnames, strict=strict)
    return _file.getvalue()


def create_tarball_on_disk(
    source_path: str,
    output_filename=None,
    compression_method="gz",
    strict=False,
):
    """
    Create a tarball of the specified file or directory.

    Parameters:
    - source_path: The path to the file or directory to add to tarball.
    - output_filename: The path where the tarball should be saved.
    """

    _paths, _arcnames = _prepare_arc_names_and_sources(source_path)
    write_string = "w"
    if compression_method is not None:
        write_string = "w:%s" % compression_method
    with tarfile.open(name=output_filename, mode=write_string) as tar:
        _save_path_and_arcs_to_tar(tar, _paths, _arcnames, strict=strict)


def _resolve_arc_name(_paths: List[str]):
    if len(_paths) == 1:
        # If we are creating a tarball when there is only ONE directory in _paths,
        # Then we want the tarball to have everything under that root directory
        # such that extracting the tarball will only the contents within that directory.
        # If we don't do this, then the tarball can have a different root directory
        # Example:
        # Assume we pass only the `path/to/my-directory` to `create_tarball` method.
        # At this point the base name of the directory is `my-directory` and if we use that
        # as the arcname, then when we extract the tarball, we will have all contents under
        # `/extraction-path/my-directory`. We don't want this since we want what ever contents
        # within `my-directory` to be directly under `/extraction-path`.

        # To avoid this, we set the arcname to an empty string such
        # that when we extract the tarball we will have all contents of `my-directory` under `/extraction-path`.
        # the os.path.basename is not a problem for files since files don't have a root directory.

        if os.path.isdir(_paths[0]):
            return [""]
    return [os.path.basename(p) for p in _paths]


def extract_tarball(tarball_path_or_bytes, target_directory, compression_method="gz"):
    """
    Extract a tarball to the specified directory.

    Parameters:
    - tarball_path_or_bytes: The path to the tarball to be extracted or the bytes of the tarball.
    - target_directory: The directory where the contents of the tarball should be extracted.
    """
    tarball_bytes_io = None
    if isinstance(tarball_path_or_bytes, bytes):
        tarball_bytes_io = BytesIO(tarball_path_or_bytes)
    else:
        tarball_path = tarball_path_or_bytes

    # Ensure the target directory exists
    if not os.path.exists(target_directory):
        os.makedirs(target_directory)

    read_string = "r"
    if compression_method is not None:
        read_string = "r:%s" % compression_method
    open_options = (
        {"fileobj": tarball_bytes_io} if tarball_bytes_io else {"name": tarball_path}
    )
    final_options = {**open_options, "mode": read_string}
    with tarfile.open(**final_options) as tar:
        tar.extractall(path=target_directory)



import os
import tempfile
from typing import Dict, List, Optional, Tuple, Union


from ..exceptions import (
    KeyNotFoundError,
    KeyNotCompatibleException,
    IncompatibleObjectTypeException,
)
from .model_storage import ModelDatastore
from .exceptions import LoadingException
from ..datastore.utils import safe_serialize
from ..utils.general import get_path_size, unit_convert, warning_message
from ..utils.identity_utils import MAX_HASH_LEN, safe_hash
from ..utils.serialization_handler import TarHandler, SERIALIZATION_HANDLERS
from ..datastructures import ModelArtifact, Factory, MetaflowDataArtifactReference
from uuid import uuid4

OBJECT_MAX_SIZE_ALLOWED_FOR_ARTIFACT = unit_convert(3, "GB", "B")


def create_write_store(pathspec, attempt, storage_backend) -> ModelDatastore:
    return ModelDatastore.init_write_store(
        storage_backend=storage_backend,
        pathspec=pathspec,
        attempt=attempt,
    )


def create_read_store(
    storage_backend,
    model_key=None,
    pathspec=None,
    attempt=None,
) -> ModelDatastore:
    return ModelDatastore.init_read_store(
        storage_backend=storage_backend,
        pathspec=pathspec,
        model_key=model_key,
        attempt=attempt,
    )


def _get_id(label):
    return "_".join([str(label), uuid4().hex.replace("-", "")])


class LoadedModels:
    """
    A class that loads models from the datastore and stores them in a temporary directory.
    This class helps manage all the models loaded via `@model(load=...)` decorator and
    `current.model.load` method.

    It is exposed via the `current.model.loaded` property. It is a dictionary like object
    that stores the loaded models in a temporary directory. The keys of the dictionary are the
    artifact names and the values are the paths to the temporary directories where the models are stored.

    Usage:
    ------
    ```python
        @model(load=["model_key", "chckpt_key"])
        @step
        def mid_step(self):
            import os
            os.listdir(current.model.loaded["model_key"])
            os.listdir(current.model.loaded["chckpt_key"])
    ```
    """

    def __init__(
        self,
        storage_backend,
        flow,
        artifact_references: Union[
            List[str],
            List[Tuple[str, Union[str, None]]],
            str,
        ],  # List of artifact names or Dict of artifact names and their subsequent paths
        best_effort=False,
        temp_dir_root=None,
        mode="eager",  # eager / lazy (not supported yet as interface needs work.)
        logger=None,
        # lazy mode will only load the model when it is accessed and not present
        # eager mode will load all the models at the start
    ) -> None:
        self._storage_backend = storage_backend
        self._loaded_model_info = {}
        self._loaded_models = {}
        self._temp_directories = {}
        self._temp_dir_root = temp_dir_root
        self._best_effort = best_effort
        # self._flow = flow
        self._logger = logger
        self._artifact_loading_mode = mode
        # self._artifact_names = artifact_names
        self._init_loaded_models(
            flow,
            artifact_references,
        )

    def _warn(self, message):
        warning_message(message, logger=self._logger, ts=False, prefix="[@model]")

    @property
    def info(self):
        return self._loaded_model_info

    def _init_loaded_models(self, flow, artifact_references):
        _art_refs = []
        if type(artifact_references) == str:
            _art_refs = [(artifact_references, None)]
        if type(artifact_references) == list:
            # Its assumed over here that the artifact_references have been validly parse until now
            for art_ref in artifact_references:
                if type(art_ref) == str:
                    _art_refs.append((art_ref, None))
                elif type(art_ref) == tuple:
                    _art_refs.append(art_ref)

        _hydrated_artifacts = []
        for artifact_name, path in _art_refs:
            artifact = getattr(flow, artifact_name, None)
            if artifact is None:
                raise LoadingException(
                    f"Artifact {artifact_name} not found in flow. Please check if `self.{artifact_name}` is defined in the flow"
                )
            if (
                type(artifact) == str
            ):  # If its a string then it means its a key reference
                try:
                    _hydrated_artifact = Factory.load_metadata_from_key(
                        artifact, self._storage_backend
                    )
                except KeyNotFoundError:
                    raise LoadingException(
                        "Artifact %s not found in the datastore" % artifact_name
                    )
                except KeyNotCompatibleException:
                    raise LoadingException(
                        "Artifact %s not compatible with any of the supported objects"
                        % artifact_name
                    )
            elif (
                isinstance(artifact, MetaflowDataArtifactReference)
                or type(artifact) == dict
            ):
                try:
                    _hydrated_artifact = Factory.hydrate(artifact)
                except ValueError:
                    raise LoadingException(
                        "Artifact %s is not a valid artifact reference. Accepted types are `str`, `dict` or `MetaflowDataArtifactReference`"
                        % artifact_name
                    )
            else:
                raise LoadingException(
                    f"Artifact %s is not a valid type. Accepted types are `str`, `dict` or `MetaflowDataArtifactReference`"
                    % artifact_name
                )
            _hydrated_artifacts.append((_hydrated_artifact, artifact_name, path))

        for (
            _hydrated_artifact,
            artifact_name,
            path,
        ) in _hydrated_artifacts:
            self._loaded_model_info[artifact_name] = _hydrated_artifact.to_dict()
            self._warn(
                "Loading Artifact with name `%s` [type:%s] with key: %s"
                % (artifact_name, _hydrated_artifact.TYPE, _hydrated_artifact.key)
            )
            self._load_artifact(artifact_name, _hydrated_artifact, path)

    def _add_model(self, artifact, path=None) -> str:
        _hydrated_artifact = None
        if type(artifact) == str:  # If its a string then it means its a key reference
            try:
                _hydrated_artifact = Factory.load_metadata_from_key(
                    artifact, self._storage_backend
                )
            except KeyNotFoundError:
                raise LoadingException(
                    f"`reference` argument string given to `current.model.load` could not be found. Reference doesn't exists in the datastore"
                )
            except KeyNotCompatibleException:
                raise LoadingException(
                    f"`reference` argument string given to `current.model.load` is not compatible with the supported artifact types"
                )
        elif (
            isinstance(artifact, MetaflowDataArtifactReference)
            or type(artifact) == dict
        ):
            try:
                _hydrated_artifact = Factory.hydrate(artifact)
            except (ValueError, IncompatibleObjectTypeException):
                raise LoadingException(
                    f"`reference` argument given to `current.model.load` is not of a valid type. Accepted types are `str`, `dict` or `MetaflowDataArtifactReference`"
                )
        else:
            raise LoadingException(
                f"`reference` argument given to `current.model.load` is not of a valid type. Accepted types are `str`, `dict` or `MetaflowDataArtifactReference`"
            )
        # Since `loaded_model_info` is user facing, we keep the object-key as the key instead
        # of the hash of the key (like we do for the temp directories)
        self._loaded_model_info[_hydrated_artifact.key] = _hydrated_artifact.to_dict()

        art_key_name = safe_hash(_hydrated_artifact.key)[:6]
        try:
            self._load_artifact(art_key_name, _hydrated_artifact, path)
        except LoadingException:
            raise LoadingException(
                f"Artifact reference specified in {_hydrated_artifact.key} not found in the datastore"
            )
        return self._loaded_models[art_key_name]

    def _load_artifact(self, artifact_name, artifact, path):
        try:
            _hydrated_artifact = Factory.hydrate(artifact)
            if path is not None:
                os.makedirs(path, exist_ok=True)
                self._loaded_models[artifact_name] = path
            else:
                self._temp_directories[artifact_name] = tempfile.TemporaryDirectory(
                    dir=self._temp_dir_root, prefix=f"metaflow_models_{artifact_name}_"
                )
                self._loaded_models[artifact_name] = self._temp_directories[
                    artifact_name
                ].name

            Factory.load(
                _hydrated_artifact,
                self._loaded_models[artifact_name],
                self._storage_backend,
            )
        except KeyNotFoundError:
            if self._best_effort:
                if artifact_name in self._temp_directories:
                    self._temp_directories[artifact_name].cleanup()

                self._loaded_models[artifact_name] = None
            raise LoadingException(
                f"Artifact reference specified in {artifact_name} not found in the datastore"
            )

    def __getitem__(self, key):
        if key not in self._loaded_models:
            raise KeyError(f"Model {key} not found in loaded models")

        return self._loaded_models[key]

    def __contains__(self, key):  # Artifact name in loaded models
        return key in self._loaded_models

    def __iter__(self):
        return iter({k: v.name for k, v in self._loaded_models.items()})

    def __len__(self):
        return len(self._loaded_models)

    def _cleanup(self):
        for name, _tempdir in self._temp_directories.items():
            if _tempdir is not None:
                print("removing tempdir for model %s at %s" % (name, _tempdir.name))
                _tempdir.cleanup()


class ModelSerializer:

    _LOADED_MODELS = None

    _saved_models = []

    def __init__(self, pathspec, attempt, storage_backend) -> None:
        self._datastore = create_write_store(pathspec, attempt, storage_backend)
        self._storage_backend = storage_backend
        self._pathspec = pathspec
        self._attempt = attempt
        flowname, _, stepname, _ = self._pathspec.split("/")
        self._default_label = "_".join([flowname, stepname])

    def _set_loaded_models(self, loaded_models: LoadedModels):
        self._LOADED_MODELS = loaded_models

    @property
    def loaded(self) -> "LoadedModels":
        return self._LOADED_MODELS

    def _get_model_artifact(
        self, metadata, serializer, storage_format, model_uuid, label=None
    ):
        return ModelArtifact.create(
            pathspec=self._pathspec,
            attempt=self._attempt,
            model_uuid=model_uuid,
            metadata=safe_serialize({} if metadata is None else metadata),
            source="task",
            serializer=serializer,
            storage_format=storage_format,
            label=label,
        )

    def _serialize_to_datastore(
        self, path, serializer: str, storage_format: str, metadata=None, label=None
    ):
        if label is None:
            label = self._default_label
        model_uuid = _get_id(label=label)
        artifact = self._get_model_artifact(
            metadata=metadata,
            serializer=serializer,
            storage_format=storage_format,
            model_uuid=model_uuid,
        )
        return self._datastore.save(artifact, path)

    def save(
        self,
        path,
        label=None,
        metadata=None,
    ):
        model_artifact = self._serialize_to_datastore(
            path,
            TarHandler.TYPE,  # TODO [POST RELEASE]: clean this up to even support directories
            TarHandler.STORAGE_FORMAT,
            metadata=metadata,
            label=label,
        ).to_dict()
        self._saved_models.append(model_artifact)
        return model_artifact

    def load(
        self,
        reference: Union[str, MetaflowDataArtifactReference, dict],
        path: Optional[str] = None,
    ):
        """
        Load a model/checkpoint from the datastore to a temporary directory or a specified path.

        Returns:
        --------
        str : The path to the temporary directory where the model is loaded.
        """

        if reference is None:
            raise ValueError(
                "reference arguement to `current.model.load` cannot be None"
            )
        # TODO [POST-RELEASE] : Implement the model_id loading
        return self._LOADED_MODELS._add_model(reference, path)


def _load_model(
    storage_backend,
    model_key=None,
    path=None,
):
    # Needed by factory methods to load the model.
    model_store = create_read_store(storage_backend, model_key=model_key)
    model_store.load(
        model_store._READ_UUID,
        path,
    )


def _load_model_metadata(
    storage_backend,
    model_key=None,
):
    # Needed by factory methods to load the model.
    model_store = create_read_store(storage_backend, model_key=model_key)
    return model_store.load_metadata(
        model_store._READ_UUID,
    )
